from typing import Any, List, Tuple
import numpy as np
# from bisect import bisect_right
from sklearn.metrics.pairwise import cosine_similarity

from corticod.algorithm.tree import Node, Tree
from corticod.algorithm.codebook import Codebook

NEW_DATA_WEIGHT = 0.25
MATURATION_ENERGY_COEFF = 10 # 50
MATURATION_ENERGY_THRESH = 100 # 500
DEFAULT_RANGE_INIT = 50
DEFAULT_RANGE_LIMIT = 10

class CorticalNode(Node):
    def __init__(self, parent=None, range_init=DEFAULT_RANGE_INIT):
        super().__init__(parent=parent)
        self.maturation_energy = 0
        self.pass_count: int = 1
        self.range = range_init
        self.range_init = range_init

    def update(self, dataIn:float, range_limit:float) -> bool:
        sqpc = np.power(self.pass_count, 0.5) # square root of pass count
        prev_data = self.get_data()
        temp_data = prev_data * (1 - NEW_DATA_WEIGHT) +  dataIn * NEW_DATA_WEIGHT 
        
        new_data = prev_data - prev_data/sqpc + temp_data/sqpc
        # new_data = (dataIn + sqpc*prev_data) / (sqpc + 1); //update node value
        self.set_data(new_data)

        self.range = self.range / (1 + self.level * sqpc)
        if self.range <= range_limit:
            self.range = range_limit # range cannot be less than range_limit

        self.pass_count += 1
        
        if not self.is_mature():
            error = abs(dataIn - new_data) + 1e-10 # to avoid division by zero
            energy_change = MATURATION_ENERGY_COEFF * self.level / error
            self.maturation_energy += energy_change
            if self.maturation_energy >= MATURATION_ENERGY_THRESH:
                self.set_mature(True)
                self.range = self.range_init
                return True # matured
        return False # no change

    def find_closest_child(self, data:float) -> Tuple[int, float]:
        
        # check only mature children first
        search_space = [np.inf if self.children.maturity_mask[i] else self.children.data_vector[i] for i in range(len(self.children.data_vector))]
        # if len(search_space) > 0:
        if sum(self.children.maturity_mask)<len(self.children.maturity_mask):
            i = np.argmin(np.abs(search_space - data))
            found_child_data = search_space[i]
            dist = abs(found_child_data - data)

            if(dist < self.children[i].range):
                return self.children[i], dist
        
        # if no mature children are close enough, check immature children
        search_space = [np.inf if not self.children.maturity_mask[i] else self.children.data_vector[i] for i in range(len(self.children.data_vector))]
        # if len(search_space) > 0:
        if sum(self.children.maturity_mask)>0:
            i = np.argmin(np.abs(search_space - data))
            found_child_data = search_space[i]
            dist = abs(found_child_data - data)

            if(dist < self.children[i].range):
                return self.children[i], dist
        
        # if no children are close enough, return None
        return None, None
    
    def find_closest_children(self, data:List[float]) -> List[int]:
        search_space = self.children.data_vector[self.children.maturity_mask]
        if len(search_space) > 0:
            dists = search_space**2 - 2 * search_space * data + data**2
            mask = dists < (self.children.range ** 2)
            return np.where(mask)[0], dists[mask]
        
        search_space = self.children.data_vector[np.bitwise_not(self.children.maturity_mask)]
        if len(search_space) > 0:
            dists = search_space**2 - 2 * search_space * data + data**2
            mask = dists < (self.children.range ** 2)
            return np.where(mask)[0], dists[mask]
        
        return [], []
        

    def get_total_progeny(self) -> int:
        total = 0
        for child in self.children:
            total += 1
            total += child.get_total_progeny()
        return total
    
    def get_level_progeny(self, target_level) -> int:
        if self.is_mature() and self.level == target_level:
            return 1
        total = 0
        for child in self.children:
            total += child.get_level_progeny(target_level)
        return total
    
    # def get_progeny_width(self) -> int:
    #     # get the widest branch along each level
    #     max_width = 0
    #     for child in self.children:
    #         width = 1 + child.get_progeny_width()
    #         if width > max_width:
    #             max_width = width
    #     return max_width


    def __repr__(self) -> str:
        log = []
        if self.parent is not None:
            log.append(f"{self.get_data():.3f}")
            log.append(f"Level: {self.level}")
        else:
            log.append("<:ROOT:>")
        log.append(f"Children: {len([c for i, c in enumerate(self.children) if self.children.maturity_mask[i]])}")
        log.append(f"Hillocks: {len([c for i, c in enumerate(self.children) if not self.children.maturity_mask[i]])}")
        log.append(f"Range: {self.range:.3f}")
        log.append(f"Pass Count: {self.pass_count}")
        # log.append(f"Maturation Energy: {self.maturation_energy:.3f}")
        log.append(f"Progeny: {self.get_total_progeny()}")
        return f"CorticalNode({', '.join(log)})"


class CortexTree(Tree):
    def __init__(self, window_size=8, range_init=DEFAULT_RANGE_INIT, range_limit=DEFAULT_RANGE_LIMIT):
        super().__init__(CorticalNode(parent=None, range_init=range_init))
        self.window_size = window_size
        self.range_limit = [range_limit * ((0.9) ** lvl)
                            for lvl in range(window_size+1)]
        self.range_init = [range_init * ((0.9) ** lvl) 
                            for lvl in range(window_size+1)]
        self.changed = True
        self.cb = None

    def closest_path(self, wave):
        node = self.root
        for coef in wave:
            if len(node.children) > 0:
                c, c_dist = node.find_closest_child(coef)
                if c:
                    if c.is_mature():
                        node = c
                        continue
                else:
                    continue
            break
        
        path = []
        while(node.parent != None):
            path.insert(0, node.get_data())
            node = node.parent
        path = np.asarray(path)
        # dist = np.linalg.norm(path - wave[:len(path)])
        # return dist if len(path) == self.window_size else 0
        return path

    def closest_full_path(self, wave):
        if self.changed:
            paths = self.paths(self.window_size)
            if paths.ndim == 1:
                return 0
            self.cb = self.complete(paths)
            self.changed = False
        return self.cb.distance(wave)

    def train_single(self, wave):
        self.changed=True
        added = 0
        leafs = 0
        node = self.root
        for coef in wave:
            if len(node.children) > 0:
                # Search among children
                found_child, c_dist = node.find_closest_child(coef)
                if found_child is not None:
                    newly_matured = found_child.update(coef, self.range_limit[found_child.level])
                    if newly_matured: 
                        # print(f"Node {found_child.get_data():.3f} matured at level {found_child.level}")
                        if found_child.level == self.window_size:
                            leafs += 1
                        added += 1
                    node = found_child
                    continue # go to next coef
            # Make new child 
            node.add_child(coef, range_init=self.range_init[node.level])
            added += 1
            break # terminate

        path = []
        while(node.parent != None):
            path.append(node.get_data())
            node = node.parent

        path = np.asarray(path)
        # dist = np.linalg.norm(path - wave[:len(path)])
        return (len(path), added, leafs)
    
    def train_epoch(self, waves):
        self.changed=True
        added = 0
        leafs = 0
        max_depth = 0
        for d_i, d in enumerate(waves):
            depth, new_added, new_leaf = self.train_single(d)
            added += new_added
            leafs += new_leaf
            max_depth = max(max_depth, depth)
            if d_i % 1000 == 0: print(f"\rProgress: {100 * (d_i/len(waves)):.2f}% - {max_depth}", end="")
        return added, leafs
    
    def train(self, waves, epochs=1):
        added = 0
        leafs = 0
        for e in range(epochs): # tqdm(range(epochs), desc="Epoch"):
            a, l = self.train_epoch(waves)
            added += a
            leafs += l
            if a > 0 or l > 0:
                print(f"\rEpoch: {e} - Added: {a}, Leaf: {l} - Codebook Length: {leafs}, Tree Size: {added}")
        return added, leafs

    def complete(self, paths=None):
        if paths is None:
            paths = self.paths(self.window_size)
        if paths.ndim == 1:
            return Codebook(np.zeros(shape=(1, self.window_size)))
        else:
            return Codebook(paths[:, :])
    


import numpy as np

class Node():
    def __init__(self, parent:'Node'=None, **kwargs):
        self.parent = parent
        self.level = parent.level + 1 if parent is not None else 0
        self.sibling_index = None
        self.children:NodeChildren = NodeChildren(self)
    def get_data(self) -> np.ndarray:
        if self.parent is None: return None
        return self.parent.children.data_vector[self.sibling_index]
    def set_data(self, data:np.ndarray) -> None:
        if self.parent is not None:
            self.parent.children.set_data(self.sibling_index, data)
    def is_mature(self) -> bool:
        if self.parent is None: return None
        return self.parent.children.maturity_mask[self.sibling_index]
    def set_mature(self, value:bool) -> None:
        if self.parent is not None: 
            self.parent.children.maturity_mask[self.sibling_index] = value
    def add_child(self, data:np.ndarray, **kwargs) -> 'Node':
        new_child = type(self)(parent=self, **kwargs)
        return self.children.add_child(new_child, data)
    def __repr__(self):
        return f"Node({self.level}:{self.get_data()})"
    
# make another class to hold all childrens of a node, and its operations
class NodeChildren():
    def __init__(self, owner:Node, dtype:np.dtype=np.float64):
        self.owner = owner
        self.count = 0
        self.elements = []
        self.data_vector = np.array([], dtype=dtype)
        self.maturity_mask = np.array([], dtype=np.bool_)

    def find_index(self, data:np.ndarray, side:str='left') -> 'Node':
        # get the first suitable index to put the data, which will not change the ascending order
        return np.searchsorted(self.data_vector, data, side=side).item()
    
    def add_child(self, new_child:Node, data:np.ndarray, maturity:bool=False) -> Node:
        i = self.find_index(data)
        new_child.sibling_index = i
        self.elements.insert(i, new_child)
        self.data_vector = np.insert(self.data_vector, i, data)
        self.maturity_mask = np.insert(self.maturity_mask, i, maturity) # True means mature, False means not mature
        for j in range(i+1, len(self.elements)):
            self.elements[j].sibling_index += 1
        self.count += 1
        return new_child
    
    def remove_child(self, i: int):
        self.count -= 1
        for j in range(i + 1, len(self.elements)): # Adjust sibling indices for elements after the removed one
            self.elements[j].sibling_index -= 1
        removed_element = self.elements.pop(i)
        removed_data = self.data_vector[i].copy()  # Copy ensures the row is not referenced
        # removed_data = self.data_vector[i, :].copy()  # Copy ensures the row is not referenced
        self.data_vector = np.delete(self.data_vector, i, axis=0)
        removed_mask = self.maturity_mask[i].copy()  # Copy for consistency
        self.maturity_mask = np.delete(self.maturity_mask, i, axis=0)
        return removed_element, removed_data, removed_mask

    def set_data(self, i:int, data:np.ndarray) -> None:
        if self.data_vector[i] == data: # no change
            return
        if i < 0 or i >= len(self.data_vector):
            print("ERROR: Child do not exist!")
            exit(-1)
        elif len(self.data_vector) == 1: # it is certain i==0 
            self.data_vector[i] = data 
        else: # There are 2 or more children
            if (i > 0 and data < self.data_vector[i - 1]) or (i < len(self.data_vector) - 1 and data > self.data_vector[i + 1]):
                # needs arrangement
                self.add_child(*self.remove_child(i))
            else: 
                self.data_vector[i] = data 
    
    def __getitem__(self, i):
        return self.elements[i]
    def __len__(self):
        return self.count
    def __iter__(self):
        return iter(self.elements)


class Tree():
    def __init__(self, root:Node):
        self.root:Node = root
        
    def paths(self, window_size:int) -> np.ndarray:
        all_paths = []
        path = []
        def recurse(node:Node):
            if len(node.children) == 0:
                if len(path) == window_size:
                    all_paths.append(path.copy())
            else:
                for child in node.children:
                    if child.is_mature():
                        path.append(child.get_data())
                        recurse(child)
                        path.pop()

        recurse(self.root)
        return np.asarray(all_paths)


import functools
import numpy as np
import torch
from sklearn.metrics.pairwise import cosine_similarity

class Util():
    @staticmethod
    def handle_list_input(original_func=None, func_2D=None, func_tensor=None):
        def _decorate(func):
            @functools.wraps(func)
            def wrapper(self, signal):
                if isinstance(signal, torch.Tensor):
                    if func_tensor is not None:
                        return func_tensor(self, signal)
                    else:
                        signal = signal.numpy()
                elif not isinstance(signal, np.ndarray):
                    signal = np.asarray(signal)
                    # if signal.ndim == 0:
                    #     signal = signal.reshape(1)

                if signal.ndim == 1 or signal.ndim == 0:
                    return func(self, signal)
                elif signal.ndim == 2:
                    if func_2D is None:
                        return np.asarray([func(self, s) for s in signal])
                    else:
                        return func_2D(self, signal)
                else:
                    raise ValueError(f"Input signal must be 1D or 2D array, got {signal.ndim}D array") 
            return wrapper

        if original_func:
            return _decorate(original_func)

        return _decorate

class Codebook():
    def __init__(self, paths, weights=None) -> None:
        if weights is not None:
            self.weights = np.asarray(weights, dtype=np.float32)
        else:
            block_size = paths[0].shape[0]
            self.weights = np.ones(block_size, dtype=np.float32)
        self.paths = paths
        self.paths_tensor = torch.tensor(self.paths).float()

    def __getitem__(self, i):
        return self.paths[i]
    
    def __len__(self):
        return len(self.paths)
    
    def __repr__(self):
        return f"Codebook with {len(self)} paths"

    @Util.handle_list_input
    def distance(self, signal):
        dist = np.linalg.norm(self.paths - signal, axis=1)
        return np.min(dist)

    @Util.handle_list_input
    def quantize(self, signal):
        return self.decode(self.encode(signal))

    @Util.handle_list_input
    def cosine_sim(self, signal):
        quantized = self.quantize(signal)
        return cosine_similarity(quantized.reshape(1, -1), signal.reshape(1, -1))[0][0]

    def encode_2d(self, signals):
        signals *= self.weights
        scaled_paths = self.paths * self.weights
        return np.argmin(np.sum(scaled_paths**2, axis=1, keepdims=True) - 2*np.dot(scaled_paths, signals.T) + np.sum(signals.T**2, axis=0, keepdims=True), axis=0)

    def encode_tensor_batch(self, signals:torch.Tensor, batch_size:int=1024*8):
        # Example of batch processing
        nearest_indices = []
        signals *= self.weights
        scaled_paths_tensor = self.paths_tensor * self.weights

        for i in range(0, signals.shape[0], batch_size):
            end = i + batch_size
            batch_distances = torch.cdist(scaled_paths_tensor, signals[i:end])
            batch_indices = torch.argmin(batch_distances, axis=0)
            # nearest_indices.append(batch_indices + i)  # Correct indices for batch offset
            nearest_indices.append(batch_indices)  # Correct indices for batch offset

        # remaining = signals.shape[0] % batch_size
        # if remaining > 0:
        #     batch_distances = torch.cdist(scaled_paths_tensor, signals[-remaining:])
        #     batch_indices = torch.argmin(batch_distances, axis=0)
        #     nearest_indices.append(batch_indices)

        nearest_indices = torch.cat(nearest_indices)
        return nearest_indices

    def encode_tensor(self, signals:torch.Tensor):
        gpu_usage, gpu_available = torch.cuda.mem_get_info()
        gpu_free = (gpu_available-gpu_usage) # in Bytes
        element_size = max(signals.element_size(), self.paths_tensor.element_size())
        signal_bytes = element_size * signals.nelement()
        path_bytes = element_size * self.paths_tensor.nelement()
        gpu_free -= signal_bytes + path_bytes
        matmul_bytes = signals.shape[0] * self.paths_tensor.shape[0] * element_size
        # if max(signal_bytes, path_bytes, matmul_bytes) > 1 * 1024 * 1024 * 1024: # > 1GB
        # print(f"Signal bytes: {signal_bytes}, Path bytes: {path_bytes}, Matmul bytes: {matmul_bytes}, GPU Free: {gpu_free}")
        if matmul_bytes > gpu_free:
            max_batch_size = int(gpu_free / (self.paths_tensor.shape[0] * element_size))
            batch_size = 2**int(np.log2(max_batch_size))
            # print(f"Decided to batch process - Batch size: {batch_size}")
            return self.encode_tensor_batch(signals, batch_size)
        else:
            signals *= self.weights
            scaled_paths_tensor = self.paths_tensor * self.weights
            return torch.argmin(torch.sum(scaled_paths_tensor**2, axis=1, keepdims=True) - 2*torch.matmul(scaled_paths_tensor, signals.T) + torch.sum(signals.T**2, axis=0, keepdims=True), axis=0)

    @Util.handle_list_input(func_2D=encode_2d, func_tensor=encode_tensor)
    def encode(self, signal):
        signal *= self.weights
        scaled_paths = self.paths * self.weights
        # print(signal, scaled_paths)
        # print(np.argmin(np.linalg.norm(scaled_paths - signal, axis=1)))
        return np.argmin(np.linalg.norm(scaled_paths - signal, axis=1))

    def decode_tensor(self, index):
        return self.paths_tensor[index]

    @Util.handle_list_input(func_tensor=decode_tensor)
    def decode(self, index):
        # print(self.paths, index)
        return self.paths[index]
    
    def encode_all(self, data):
        encoded = []
        for w in data:
            encoded.append(self.encode(w))
        encoded = np.asarray(encoded)


if __name__ == "__main__":

    # Example paths for the codebook (2D arrays)
    # paths = [
    #     np.array([[1.0, 0.0], [0.0, 1.0]]),
    #     np.array([[0.0, 1.0], [1.0, 0.0]]),
    #     np.array([[1.0, 1.0], [0.0, 0.0]]),
    #     np.array([[0.5, 0.5], [0.5, 0.5]])
    # ]

    # # Example paths for the codebook
    paths = [
        np.array([1.0, 0.0, 0.0]),
        np.array([0.0, 1.0, 0.0]),
        np.array([0.0, 0.0, 1.0]),
        np.array([1.0, 1.0, 1.0])
    ]

    # Initialize Codebook
    codebook = Codebook(paths)

    # # Example signals for testing
    signals = np.array([
        [0.9, 0.1, 0.0],
        [0.0, 0.8, 0.2],
        [0.3, 0.3, 0.4],
        [0.99, 0.99, 0.99]
    ])

    # Example 2D signals for testing
    # signals = [
    #     np.array([[0.9, 0.1], [0.1, 0.9]]),
    #     np.array([[0.0, 0.8], [0.2, 0.0]]),
    #     np.array([[0.3, 0.3], [0.4, 0.4]]),
    #     np.array([[1.0, 1.0], [0.0, 0.0]])
    # ]

    print("Testing Codebook Operations:\n")

    # Test distance function
    for signal in signals:
        dist = codebook.distance(signal)
        print(f"Distance for {signal}: {dist}")

    # Test quantize function
    for signal in signals:
        quantized_signal = codebook.quantize(signal)
        print(f"Quantized {signal} to {quantized_signal}")

    # Test cosine similarity
    for signal in signals:
        similarity = codebook.cosine_sim(signal)
        print(f"Cosine similarity for {signal}: {similarity}")

    # Test encode and decode
    for signal in signals:
        encoded_index = codebook.encode(signal)
        decoded_signal = codebook.decode(encoded_index)
        print(f"Encoded {signal} to index {encoded_index}, Decoded to {decoded_signal}")

    # Test tensor operations (batch encoding)
    signals_tensor = torch.tensor(signals).float()
    encoded_indices = codebook.encode(signals_tensor)
    decoded_signals = codebook.decode(encoded_indices)

    print("\nTensor Operations:")
    print(f"Encoded Indices: {encoded_indices}")
    print(f"Decoded Signals: {decoded_signals}")




from setuptools import setup, find_packages

setup(
    name="corticod",
    version="0.1.1",
    description="A biomimetic cortical coding library for audio compression and signal processing.",
    author="Ahmet Emin Ünal",
    author_email="aeunal@hotmail.com",
    url="https://github.com/itu-lab/CorticalCodingCodec", 
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",  
    packages=find_packages(),
    package_dir= {
        'corticod': 'corticod',
        'corticod.algorithm': 'corticod/algorithm',
        'corticod.utils': 'corticod/utils'
    },
    install_requires=[
        "numpy>=2.0.2",
        "pandas>=2.2.3",
        "scikit-learn>=1.6.0",
        "scikit-image>=0.24.0",
        "soundfile>=0.12.1",
        "path",
        "torch",
        "torchvision",
        "torchaudio",
        "torchmetrics>=1.6.1",
        "scikit-image",
        "constriction"
    ],
    dependency_links=[
        'https://download.pytorch.org/whl/cu118/'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    include_package_data=True  # Includes non-code files like README
)
